' Custom-built OpenFile dialog via template in resource.
' Public Domain by Borje Hagsten, December 2002
' Free to use and modify - but as always, at own responsibility.
'
' Comments, questions or bug-reports, etc. can be mailed to me at
' pb@tolkenxp.com. No guarantees for quick response, but I'll do
' my best to keep up with eventual mail. Note - use the latest
' include files from PB with this one, at least latest Comdlg32.inc
'
' LOG:
' Oct 17 -04: Changed from using CONTROL ADD to more correct CreateWindowEx
'             in FUNCTION ofnBuildTabControl. CONTROL ADD should only be
'             used with DDT dialogs, not system created dialogs like OpenFile..

' Mar 25: Updated some ofnOpenFileTemplate code for PB/WIN 7.01
'         changes to ACIIZ string behaviour with embedded nuls.
' Feb 17: Fixed a potential "eternal loop" in ofnIniToListView by
'         adding a second, independent variable, d, for ofnIniGetString call.
' Dec 22: Added listview column re-adjustment under WM_SIZE in ofnHookProc.
'
' Dec 19: Rewrote resizing code a bit so it handles changes in resource
'         dialog size and LTEXT position label better. Added some
'         explanations about size/pos to resource file, Ofn.rc.
'         But resize is tricky, so if you want to add controls, you
'         may have to experiment and adjust things under WM_SIZE
'
' Dec 18: Added %WM_SYSCOLORCHANGE forwarding from ofnComDlgProc to
'         TabMain and ListView. Also added font-update code for Tab
'         under same message, so eventual color scheme -changes are
'         handled directly if that happens when dialog is shown.
'
' Dec 18: Fixed a mix-up with OPENFILENAME5 vs. OPENFILENAME in code.
'         Added a global string for results from lv in tab page 2. Some
'         have reported problems in Win2K/XP, so maybe these systems move
'         memory around, and an address grabbed at load may not be valid
'         at exit. Hard to say without being able to test, but using a
'         global string instead should be 100% ok. Could thereby replace
'         OPENFILENAME pointer in structure with ofnFlags DWORD.
'         Added ofnFileExist check to ofnIniToListView routine.
'         Also cleaned out a bunch of un-used locals, found by PBcodec..

' Dec 17: Removed SetParent code for Ok button and moved trap for it
'         from ofnTabPage2Proc into ofnComDlgProc. Looks better and saved some code.
'         Renamed all routines belonging to the dialog to start with ofn.
'         Placed most routines in sorted order, but test dialog code first.
'         Added UpdateWindow after control resize in OfnHookProc's WM_SIZE,
'         for quicker update = less flicker (tab control resize can be nasty..)
'
' Dec 17: Added %WS_CLIPSIBLINGS style to tab dialog
'         Added redraw for ImageLabel under WM_PAINT in OfnHookProc
'         Removed some InvalidateRects from WM_SIZE in OfnHookProc
'         Added ofnIsWin2000orLater routine + call in ofnOpenFileTemplate
'         Changed a few variables from DWORD to LONG, to match declares
'         Added ofnFileExist checks for resulting file names in MyOpenFileProc
'
' INTRO:
' Built with "professional use" in mind and thoroughly tested
' under Win95/98. No guarantees for later systems, but reports
' are welcome. Dialog can handle resize and multiple selection,
' and takes eventual help/read-only controls into consideration.
' Routines for storing/loading a Recent files -list in/from ini
' file is here too. Please feel free to modify to your own needs.
'
' Sample provides several ways to test difference OpenFile flags.
' Select files in OpenFile and press "Open" (or double-click to open),
' and the file(s) will be added to "Recent" section in OfnTempl.ini.
' Open the dialog again, and they'll be in the Recent-files ListView.
'
' BASICS:
' The OpenFile dialog is hook'd and hookproc is told to use a template
' dialog from resource. This resource template dialog decides how much
' additional width/height the OpenFile dialog shall have.
'
' In the template dialog exist only one control, a label with id stc32.
' This label decides position for OpenFile dialog in template. ComDlg
' knows what this label is meant for - MS has told. Also see Ofn.rc (resource file)
'
' USAGE:
' PBMAIN and DlgProc is sample program. MyOpenFileProc is like in-between
' CommonDlg template code and sample prog, and since it includes code for
' handling multiple files selection, it can be worth looking at. Ini file
' routines are also a bit in-between, but is needed for Recent files list.
' The rest is needed for hook'd and subclassed ComDlg. Look for ofn in the
' beginning Should not be too hard to integrate this one in other projects,
' I think/hope.
'
' COMMENTS:
' Some tricky stuff here. The OpenFile dialog is parent to the template,
' but still the template visually looks like parent to OpenFile. So
' we use OpenFile as parent to tab control, and still it looks like
' OpenFile is child of tab control. Yeah, it can take a while to digest,
' believe me - I know.. :)
'
' Since OpenFile is sort of a black box, some trix must be used. Instead
' of using different OK buttons on different tab pages, I switch parent
' for it when tab pages are changed. Need some reposition to make it look
' like pos is same, but no problems - see OfnHookProc, under %WM_NOTIFY.
'
' Also found that I had to subclass the darn thing, because WM_COMMAND
' and WM_SYSCOMMAND never reaches hook-procedure, and we must consider
' cancel actions from sys-menu and Cancel button when Recent files tab
' is active. A very small sub-class proc (ofnComDlgProc), so no big deal..
' 
'
' Declares - first eliminate some stuff from commctrl.inc
'
%NOANIMATE         = 1  ' Animate control
%NOBUTTON          = 1  ' Button
%NOCOMBO           = 1  ' Combo box
%NOCOMBOEX         = 1  ' ComboBoxEx
%NODATETIMEPICK    = 1  ' Date/time picker
%NODRAGLIST        = 1  ' Drag list control
%NOEDIT            = 1  ' Edit control
%NOFLATSBAPIS      = 1  ' Flat scroll bar
%NOHEADER          = 1  ' Header control
%NOHOTKEY          = 1  ' HotKey control
'%NOIMAGELIST       = 1  ' Image APIs
%NOIPADDRESS       = 1  ' IP Address edit control
%NOLIST            = 1  ' List box control
'%NOLISTVIEW        = 1  ' ListView control
%NOMENUHELP        = 1  ' Menu help
%NOMONTHCAL        = 1  ' MonthCal
%NOMUI             = 1  ' MUI
%NONATIVEFONTCTL   = 1  ' Native Font control
%NOPAGESCROLLER    = 1  ' Pager
%NOPROGRESS        = 1  ' Progress control
%NOREBAR           = 1  ' Rebar control
%NOSTATUSBAR       = 1  ' Status bar
'%NOTABCONTROL      = 1  ' Tab control
%NOTOOLBAR         = 1  ' Tool bar
%NOTOOLTIPS        = 1  ' Tool tips
%NOTRACKBAR        = 1  ' Track bar
%NOTRACKMOUSEEVENT = 1  ' Track Mouse Event
%NOTREEVIEW        = 1  ' TreeView
%NOUPDOWN          = 1  ' Up Down arrow control
'--------------------------------------------------------------------
#COMPILE EXE
#INCLUDE "WIN32API.INC"
#INCLUDE "COMMCTRL.INC"
#INCLUDE "COMDLG32.INC"
#RESOURCE "OFN.PBR"      '<- contains template dialog
'--------------------------------------------------------------------
%stc32         = 1119             'id for label in template dialog - don't change it!
%IDC_IMG       = %WM_USER + 2048  'id for "image" label
%IDC_TAB       = %WM_USER + 2049  'id for tab control
%IDC_LISTVIEW2 = %WM_USER + 2050  'id for listview in tab page 2
'--------------------------------------------------------------------
DECLARE CALLBACK FUNCTION DlgProc() AS LONG
DECLARE FUNCTION MyOpenFileProc (BYVAL hDlg AS DWORD, BYVAL Path AS STRING) AS STRING
'--------------------------------------------------------------------
DECLARE FUNCTION ofnBuildTabControl(BYVAL hWnd AS DWORD) AS DWORD
DECLARE FUNCTION ofnComDlgProc(BYVAL hWnd AS DWORD, BYVAL wMsg AS DWORD, _
                               BYVAL wParam AS DWORD, BYVAL lParam AS LONG) AS DWORD
DECLARE FUNCTION ofnEnumChildsProc(BYVAL hWnd AS DWORD, BYVAL lParam AS DWORD) AS DWORD
DECLARE FUNCTION ofnFileExist(BYVAL sfName AS STRING) AS LONG
DECLARE FUNCTION ofnGetFileNamePart(BYVAL Src AS STRING) AS STRING
DECLARE FUNCTION ofnGetFilePathPart(BYVAL Src AS STRING) AS STRING
DECLARE FUNCTION ofnIniGetString(BYVAL sSection AS STRING, BYVAL sKey AS STRING, _
                              BYVAL sDefault AS STRING) AS STRING
DECLARE FUNCTION ofnIniSetString(BYVAL sSection AS STRING, BYVAL sKey AS STRING, _
                                 BYVAL sStr AS STRING) AS LONG
DECLARE FUNCTION ofnIsWin2000orLater AS LONG
DECLARE FUNCTION ofnMakeFont(BYVAL sFont AS STRING, BYVAL PointSize AS LONG, _
                             BYVAL fBold AS LONG, BYVAL fItalic AS LONG, _
                             BYVAL fUnderline AS LONG, BYVAL StrikeThru AS LONG) AS LONG
DECLARE FUNCTION ofnMakeFontEnumCharSet (elf AS ENUMLOGFONT, ntm AS NEWTEXTMETRIC, _
                                         BYVAL FontType AS LONG, CharSet AS LONG) AS LONG
DECLARE FUNCTION ofnOpenFileTemplate  (BYVAL hWnd AS DWORD, _           ' parent window
                                       BYVAL Caption AS STRING, _       ' caption
                                       Filespec AS STRING, _            ' filename
                                       BYVAL InitialDir AS STRING, _    ' start directory
                                       BYVAL Filter AS STRING, _        ' filename filter
                                       BYVAL DefExtension AS STRING, _  ' default extension
                                       Flags AS DWORD) AS LONG          ' flags
DECLARE FUNCTION OfnHookProc(BYVAL hWnd AS DWORD, BYVAL wMsg AS DWORD, _
                             BYVAL wParam AS DWORD, BYVAL lParam AS LONG) AS DWORD

DECLARE CALLBACK FUNCTION ofnTabPage2Proc() AS LONG

DECLARE SUB ofnCenterChild(BYVAL hMain AS DWORD, BYVAL hChild AS DWORD)
DECLARE SUB ofnIniToListView
DECLARE SUB ofnWriteRecentFiles (BYVAL OpenFName AS STRING)
'--------------------------------------------------------------------
TYPE OFNTEMPLATEDATA
  oldofnComDlgProc AS DWORD
  hImgLabel        AS DWORD
  hListView        AS DWORD
  hTabmain         AS DWORD
  hFont            AS DWORD
  ofnFlags         AS DWORD
  TabPg            AS LONG
  hTab(1)          AS DWORD
  zIni             AS ASCIIZ * %MAX_PATH
END TYPE
                               ' a few globals makes life easier here..
GLOBAL gOfnFiles AS STRING     ' for storing files from tab page 2
GLOBAL otd AS OFNTEMPLATEDATA  ' see UDT above
GLOBAL ofnCtrl() AS DWORD      ' for storing ComClg control handles

'
' Main entrance
'
FUNCTION PBMAIN () AS LONG
  LOCAL hDlg AS DWORD

  GetModuleFileName GetModuleHandle(""), otd.zIni, %MAX_PATH 'get part to prog.
  otd.zIni = ofnGetFilePathPart(otd.zIni) + "OfnTempl.ini"   'extract path and add ini file

  DIALOG NEW 0, "OpenFile template..",,, 112, 92, _          'build test dialog and some controls
                %WS_CAPTION OR %WS_SYSMENU, 0 TO hDlg

  CONTROL ADD BUTTON, hDlg, %IDOK,     "&Open",  5, 74, 50, 14
  CONTROL ADD BUTTON, hDlg, %IDCANCEL, "&Quit", 57, 74, 50, 14

  CONTROL ADD FRAME,    hDlg, 10, "OpenFile's settings",   6,  4, 100, 63
  CONTROL SET COLOR hDlg, 10, RGB(128, 0, 0), -1

  CONTROL ADD CHECKBOX, hDlg, 40, "Show &Help button ",   12, 16,  90, 10
  CONTROL ADD CHECKBOX, hDlg, 41, "&Read-only checkbox ", 12, 28,  90, 10
  CONTROL ADD CHECKBOX, hDlg, 42, "Re&sizable dialog ",   12, 40,  90, 10
  CONTROL ADD CHECKBOX, hDlg, 43, "Allow &multi-select ", 12, 52,  90, 10

  DIALOG SHOW MODAL hDlg CALL DlgProc

END FUNCTION

'
' Main dialog's callback proc
'
CALLBACK FUNCTION DlgProc() AS LONG

  SELECT CASE CBMSG
     CASE %WM_COMMAND
        SELECT CASE CBCTL
           CASE %IDOK
              IF CBCTLMSG = %BN_CLICKED THEN
                 CALL MyOpenFileProc(CBHNDL, "")
              END IF

           CASE %IDCANCEL
              IF CBCTLMSG = %BN_CLICKED THEN DIALOG END CBHNDL
        END SELECT

  END SELECT
END FUNCTION

'
' Show OpenFile dialog and handle returned file names, if any
'
FUNCTION MyOpenFileProc (BYVAL hDlg AS DWORD, BYVAL Path AS STRING) AS STRING
  LOCAL c AS LONG, fCount AS LONG, dwStyle AS DWORD
  LOCAL sFile AS STRING, sBuf AS STRING, fName AS STRING
   
  IF LEN(Path) = 0 THEN Path = CURDIR$ 'if no path was given..
  fName   = "*.*"
  sBuf    = "All Files (*.*)|*.*|Text Files (*.TXT)|*.TXT"
  dwStyle = %OFN_EXPLORER OR %OFN_ENABLEHOOK OR %OFN_ENABLETEMPLATE OR _
            %OFN_FILEMUSTEXIST OR %OFN_PATHMUSTEXIST

  CONTROL GET CHECK hDlg, 40 TO c                   ' checkboxes - what did you select?
  IF c = 1 THEN dwStyle = dwStyle OR %OFN_SHOWHELP

  CONTROL GET CHECK hDlg, 41 TO c
  IF c = 0 THEN dwStyle = dwStyle OR %OFN_HIDEREADONLY

  CONTROL GET CHECK hDlg, 42 TO c
  IF c = 1 THEN dwStyle = dwStyle OR %OFN_ENABLESIZING

  CONTROL GET CHECK hDlg, 43 TO c
  IF c = 1 THEN dwStyle = dwStyle OR %OFN_ALLOWMULTISELECT

  IF ofnOpenFileTemplate(hDlg, "", fName, Path, sBuf, "TXT", dwStyle) THEN 'call OpenFile dialog

     IF INSTR(fName, "|") THEN            'if "|", then it comes from RECENT FILES list..
        fCount = PARSECOUNT(fName, "|")   'parse through string, in case of multiple files
        FOR c = 1 TO fCount
           sBuf = PARSE$(fName, "|", c)        ' parse out file
           IF ofnFileExist(sBuf) THEN          ' make sure it exists
             'open file -code can be placed here..
              sFile = sFile + sBuf + $CRLF     ' full path and file name
              ofnWriteRecentFiles sBuf         ' update Recent files list
           END IF
        NEXT

     ELSE                                      ' else if comes from OPENFILE DIALOG
        fCount = PARSECOUNT(fName, CHR$(0))    ' see how many files we got
        IF fCount = 1 THEN                     ' if a single file
           sFile = PARSE$(fName, CHR$(0), 1)   ' full path and file name in same
           IF ofnFileExist(sFile) THEN         ' make sure it really exists
             'open file -code can be placed here..
              ofnWriteRecentFiles sFile           ' update Recent files list
           END IF

        ELSE                                   ' if multiple files
           Path = PARSE$(fName, CHR$(0), 1)    ' ComDlg returns path first, then names
           IF RIGHT$(Path, 1) <> "\" THEN Path = Path & "\" 'make sure it's ok
           FOR c = 2 TO fCount                              'then parse for file names
              sBuf = Path + PARSE$(fName, CHR$(0), c)
              IF ofnFileExist(sBuf) THEN       ' make sure it exists
                'open file -code can be placed here..
                 sFile = sFile + sBuf + $CRLF
                 ofnWriteRecentFiles sBuf
              END IF
           NEXT
        END IF
     END IF

     IF (dwStyle AND %OFN_READONLY) OR _     ' if read-only checkbox was marked
        (GETATTR(fName) AND %READONLY) THEN  ' or file has read-only attribute set
                                             ' here should OPEN, etc, follow..

        MSGBOX "File(s) to open as read-only: " + $CRLF + $CRLF + sFile
     ELSE
        MSGBOX "File(s) to open for editing: " + $CRLF + $CRLF + sFile
     END IF

  END IF
END FUNCTION

 
'
' Build and show tab control
'
FUNCTION ofnBuildTabControl(BYVAL hWnd AS DWORD) AS DWORD
  LOCAL dwStyle AS DWORD, hTabMain AS DWORD, txt AS STRING, rc AS RECT
  LOCAL lvc AS LV_COLUMN, ncm AS NONCLIENTMETRICS, tci AS TC_ITEM

  InitCommonControls                ' Initialize the common control library
  otd.hTab(0) = GetParent(hWnd)     ' let otd.hTab(0) hold comdlg's handle

  otd.hImgLabel = CreateWindowEx(0, "STATIC", "", _
                                 %WS_CHILD OR %WS_VISIBLE OR %SS_SUNKEN, _
                                 7, 7, 217, 22, _
                                 hWnd, %IDC_IMG, GetModuleHandle(""), BYVAL 0)

  otd.hTabMain = CreateWindowEx(0, "SysTabControl32", "", _
                                 %WS_CHILD OR %WS_TABSTOP OR %WS_VISIBLE OR %WS_CLIPSIBLINGS, _
                                 6, 29, 216, 136, _
                                 hWnd, %IDC_TAB, GetModuleHandle(""), BYVAL 0)

  ncm.cbSize = SIZEOF(ncm) 'grab a good system font for tab control
  SystemParametersInfo %SPI_GETNONCLIENTMETRICS, SIZEOF(ncm), BYVAL VARPTR(ncm), 0
  IF LEN(ncm.lfStatusFont) THEN otd.hFont = CreateFontIndirect(ncm.lfStatusFont)
  IF otd.hFont THEN
     SendMessage otd.hTabMain, %WM_SETFONT, otd.hFont, 0
  ELSE
     SendMessage otd.hTabMain, %WM_SETFONT, GetStockObject(%ANSI_VAR_FONT), 0
  END IF

  GetClientRect otd.hTab(0), rc 'grab an intial size - WM_SIZE will adjust size later on..
  DIALOG PIXELS otd.hTab(0), rc.nRight, rc.nBottom TO UNITS rc.nRight, rc.nBottom

 
  'create tab page 1 (&Open) Note: no dialog needed here - we use comdlg for this one
  txt = SPACE$(100)
  GetWindowText GetDlgItem(otd.hTab(0), %IDOK), _
                BYVAL STRPTR(txt), LEN(txt)  ' maybe we can use Ok button's text?
  IF LEN(txt) = 0 THEN                       ' if not, use own caption
     txt = " &Open "
  ELSE
     txt = " " + txt + " "
  END IF
  tci.mask    = %TCIF_TEXT
  tci.pszText = STRPTR(txt)
  TabCtrl_InsertItem otd.hTabMain, 0, tci

  'create tab page 2 (&Recent)
  txt = " &Recent "  ' if you want to localize, "&Recent" is the only one you need to translate..
  tci.pszText = STRPTR(txt)
  TabCtrl_InsertItem otd.hTabMain, 1, tci
  DIALOG NEW otd.hTab(0), "" , 0, 0, rc.nRight + 1, rc.nBottom - 15, _
                               %WS_CHILD OR %DS_CONTROL OR %WS_CLIPSIBLINGS, 0 TO otd.hTab(1)

  'add controls to tab page 2
  dwStyle = %WS_CHILD OR %WS_VISIBLE OR %WS_VSCROLL OR _
            %WS_HSCROLL OR %WS_TABSTOP OR %LVS_REPORT OR _
            %LVS_SHAREIMAGELISTS OR %LVS_SHOWSELALWAYS

  IF (otd.ofnFlags AND %OFN_ALLOWMULTISELECT) = 0 THEN ' if OpenFile wasn't told to allow
     dwStyle = dwStyle OR %LVS_SINGLESEL               ' multiple seletions (which is ListView def.)
  END IF

  otd.hListView = CreateWindowEx(%WS_EX_CLIENTEDGE, "SysListView32", "", _
                                 dwStyle, _
                                 8, 8, rc.nRight - 8, 100, _
                                 otd.hTab(1), %IDC_LISTVIEW2, GetModuleHandle(""), BYVAL 0)

  dwStyle = SendMessage(otd.hListView, %LVM_GETEXTENDEDLISTVIEWSTYLE, 0, 0) ' add full-row select and
  dwStyle = dwStyle OR %LVS_EX_FULLROWSELECT OR %LVS_EX_INFOTIP             ' tooltip when mouse is over
  SendMessage otd.hListView, %LVM_SETEXTENDEDLISTVIEWSTYLE, 0, dwStyle      ' text that is too long

  'setup listview on page 2
    GetClientRect otd.hListView, rc 'columns will be re-adjusted under WM_SIZE..
    lvc.mask    = %LVCF_FMT OR %LVCF_TEXT OR %LVCF_WIDTH OR %LVCF_SUBITEM
    lvc.fmt     = %LVCFMT_LEFT

    txt         = "File"           ' first column - file names
    lvc.pszText = STRPTR(txt)
    lvc.cx      = rc.nRight * 0.3  ' 30% of listview width for file name
    lvc.iOrder  = 0
    SendMessage otd.hListView, %LVM_INSERTCOLUMN, 0, VARPTR(lvc)

    txt         = "Folder"         ' second column - folder names
    lvc.pszText = STRPTR(txt)
    lvc.cx      = rc.nRight * 0.7  ' 70% of total width for path
    lvc.iOrder  = 1
    SendMessage otd.hListView, %LVM_INSERTCOLUMN, 1, VARPTR(lvc)

 
  DIALOG SHOW MODELESS otd.hTab(1), CALL ofnTabPage2Proc
  DIALOG SHOW STATE otd.hTab(1), %SW_HIDE 'hide tab2 dialog initially
  TabCtrl_SetCurFocus otd.hTabMain, 0     'give tab1 focus

END FUNCTION

'
' center a child dialog above/in parent
'
SUB ofnCenterChild(BYVAL hMain AS DWORD, BYVAL hChild AS DWORD)
  LOCAL rc AS RECT, x AS LONG, y AS LONG, w AS LONG, h AS LONG

  GetWindowRect hMain, rc    ' get parent's pos. and size
  w = rc.nRight  - rc.nLeft  ' calculate width
  h = rc.nBottom - rc.nTop   ' calculate height
  x = rc.nLeft + (w \ 2)     ' relative Center x
  y = rc.nTop  + (h \ 2)     ' relative Center y

  GetWindowRect hChild, rc   ' get childs pos. and size
  w = rc.nRight  - rc.nLeft  ' width
  h = rc.nBottom - rc.nTop   ' height
  x = x - (w \ 2)            ' centered x pos. in/above program
  y = y - (h \ 2)            ' centered y pos. in/above program

  SystemParametersInfo %SPI_GETWORKAREA, BYVAL %NULL, BYVAL VARPTR(rc), 0 ' get desktop size
  IF x < rc.nLeft THEN x = rc.nLeft  ' make sure it's on screen
  IF x + w > rc.nRight THEN x = rc.nRight - w
  IF y < rc.nTop THEN y = rc.nTop
  IF y + h > rc.nBottom THEN y = rc.nBottom - h

  SetWindowPos hChild, 0, x, y, 0, 0, %SWP_NOSIZE OR %SWP_NOZORDER 'time to set centered position

END SUB

'
' Must sub-class comdlg, because %WM_COMMAND and %WM_SYSCOMMAND are
' never sent to hook'd procedure. So we can handle cancel properly..
'
FUNCTION ofnComDlgProc(BYVAL hWnd AS DWORD, BYVAL wMsg AS DWORD, _
                       BYVAL wParam AS DWORD, BYVAL lParam AS LONG) AS DWORD

  LOCAL i AS LONG, j AS LONG, ncm AS NONCLIENTMETRICS
  LOCAL zfName AS ASCIIZ * %MAX_PATH, zPath AS ASCIIZ * %MAX_PATH

  SELECT CASE wMsg
     CASE %WM_COMMAND      ' Cancel or Escape key pressed
        IF LOWRD(wParam) = %IDCANCEL THEN
           otd.TabPg = 0

        ELSEIF LOWRD(wParam) = %IDOK THEN
           IF TabCtrl_GetCurSel(otd.hTabMain) > 0 THEN  ' if not tab page 0, set Ok button's pos
              gOfnFiles = ""                            ' ensure global string is empty (should be, but..)

              FOR i = 0 TO ListView_GetItemCount(otd.hListView)              'loop though all items
                 j = ListView_GetItemState(otd.hListView, i, %LVIS_SELECTED) 'get sel status
                 IF (j AND %LVIS_SELECTED) = %LVIS_SELECTED THEN             'if selected
                    ListView_GetItemText otd.hListView, i, 0, zfName, SIZEOF(zfName) 'grab file name
                    ListView_GetItemText otd.hListView, i, 1, zPath, SIZEOF(zPath)   'grab path
                    IF ofnFileExist (zPath & zfName) THEN                    'make sure it exists
                       gOfnFiles = gOfnFiles + zPath + zfName + "|"          'add to result string        
                    END IF
                    IF (otd.ofnFlags AND %OFN_ALLOWMULTISELECT) = 0 THEN EXIT FOR 'is single sel only
                 END IF
              NEXT

              i = GetDlgItem(otd.hTab(0), %chx1)                   ' look for read-only checkbox
              IF i AND IsWindowEnabled(i) THEN                     ' if it's there and enabled
                 IF SendMessage(i, %BM_GETCHECK, 0, 0) THEN        ' see if it's checked
                    otd.ofnFlags = otd.ofnFlags OR %OFN_READONLY  ' set read-only flag for return
                 END IF
              END IF

              IF LEN(gOfnFiles) THEN               ' if we have anything to return
                 EndDialog otd.hTab(0), 0          ' close dialog

              ELSE                                          ' Common dialog redraws dialog if Ok is
                 InvalidateRect otd.hTab(1), BYVAL %NULL, 0 ' pressed and nothing is selected. Is a 
                 UpdateWindow otd.hTab(1)                   ' smart way to notify user, so we do same..
              END IF

              EXIT FUNCTION
           END IF
        END IF

     CASE %WM_SYSCOLORCHANGE 'color-scheme change - must forward this one to common controls
        SendMessage otd.hTabMain, wMsg, wParam, lParam
        SendMessage otd.hListView, wMsg, wParam, lParam

        ncm.cbSize = SIZEOF(ncm) ' need to update font in tab control too
        SystemParametersInfo %SPI_GETNONCLIENTMETRICS, SIZEOF(ncm), BYVAL VARPTR(ncm), 0
        IF LEN(ncm.lfStatusFont) THEN
           IF otd.hFont THEN DeleteObject otd.hFont            ' delete previously created font, if any
           otd.hFont = CreateFontIndirect(ncm.lfStatusFont)    ' create a new one
        END IF
        IF otd.hFont THEN
           SendMessage otd.hTabMain, %WM_SETFONT, otd.hFont, 0 ' and set new one to tab control
        ELSE
           SendMessage otd.hTabMain, %WM_SETFONT, GetStockObject(%ANSI_VAR_FONT), 0
        END IF

     CASE %WM_SYSCOMMAND   ' see if user closed from sysmenu
        IF wParam = %SC_CLOSE THEN otd.TabPg = 0

  END SELECT
  FUNCTION = CallWindowProc(otd.oldofnComDlgProc, hWnd, wMsg, wParam, lParam)
END FUNCTION

'
' Enumerate and store needed OpenFile dialog child hWnd's in an array
'
FUNCTION ofnEnumChildsProc(BYVAL hWnd AS DWORD, BYVAL lParam AS DWORD) AS DWORD
  LOCAL az AS ASCIIZ * 80

  GetClassName hWnd, az, SIZEOF(az) ' see what class child's handle belongs to
  az = UCASE$(az)                   ' UCASE for compare

  IF az <> "#32770" AND _  'filter out what we don't want to hide/show - like dialogs
     az <> "SYSTABCONTROL32" AND _        ' ignore tab control
     az <> "BUTTON" AND _                 ' ignore buttons
     az <> "LISTBOX" AND _                ' ignore old style's listbox
     az <> "SCROLLBAR" AND _              ' scrollbar? yes - in resizable dialog..
     GetDlgCtrlId(hWnd) <> %IDC_IMG AND _ ' our dear image label must stay
     isWindowEnabled(hWnd) AND _          ' only enabled childs are interesting
     isWindowVisible(hWnd) THEN           ' and visible ones, of course

     ofnCtrl(UBOUND(ofnCtrl)) = hWnd      ' if it passed all that, store handle in array because
     REDIM PRESERVE ofnCtrl(UBOUND(ofnCtrl) + 1) ' we'll need to hide/show it on tab page change
  END IF

  FUNCTION = 1
END FUNCTION

'
' ofnFileExist - make sure a file or folder exists
'
FUNCTION ofnFileExist(BYVAL sfName AS STRING) AS LONG
  LOCAL hRes AS DWORD, tWFD AS WIN32_FIND_DATA

  IF LEN(sfName) = 0 THEN EXIT FUNCTION       'no need to continue then..
  IF ASC(sfName, -1) = 92 THEN                'if a path with trailing backslash
     sfName = LEFT$(sfName, LEN(sfName) - 1)  'remove trailing backslash
  END IF

  hRes = FindFirstFile(BYVAL STRPTR(sfName), tWFD)
  IF hRes <> %INVALID_HANDLE_VALUE THEN
     FUNCTION = %TRUE
     FindClose hRes
  END IF
END FUNCTION

'
' Extract and return file name -part
'
FUNCTION ofnGetFileNamePart(BYVAL Src AS STRING) AS STRING
  LOCAL nPos AS LONG

  nPos = INSTR(-1, Src, ANY "\/")
  IF nPos THEN
     FUNCTION = MID$(Src, nPos + 1)
  ELSE
     FUNCTION = Src
  END IF
END FUNCTION

'
' Extract and return file path -part
'
FUNCTION ofnGetFilePathPart(BYVAL Src AS STRING) AS STRING
  LOCAL nPos AS LONG

  nPos = INSTR(-1, Src, ANY "\/")
  IF nPos THEN
     FUNCTION = LEFT$(Src, nPos)
  ELSE
     FUNCTION = Src
  END IF
END FUNCTION

'
' Template procedure, which becomes main OpenFile dialog proc.
' Note: hWnd belongs to template dialog - hWnd's parent is comdlg..
' OfnHookProc actually becomes Tab page 1's procedure.
'
FUNCTION OfnHookProc(BYVAL hWnd AS DWORD, BYVAL wMsg AS DWORD, _
                     BYVAL wParam AS DWORD, BYVAL lParam AS LONG) AS DWORD

  LOCAL ii AS LONG, w AS LONG, h AS LONG, lPage AS LONG, hftCombo AS DWORD
  LOCAL pNhmdr AS NMHDR PTR, rc AS RECT, rc2 AS RECT, zTxt AS ASCIIZ * 100

  SELECT CASE wMsg
     CASE %WM_INITDIALOG
        MOUSEPTR 11       ' set hour-glass mouse pointer (comdlg resets to arrow when load is ready)
        ShowWindow GetDlgItem(hWnd, %stc32), %SW_HIDE    ' hide position-label in template
        ofnBuildTabControl hWnd                          ' build tab control
        ofnIniToListView                                 ' fill ListView from ini file
        otd.oldofnComDlgProc = SetWindowLong(GetParent(hWnd), _  ' subclass comdlg
                               %GWL_WNDPROC, CODEPTR(ofnComDlgProc))


     CASE %WM_CTLCOLORSTATIC                          ' this is instead of bitmap in resource..
        IF GetDlgCtrlId(lParam) = %IDC_IMG THEN       ' make sure it's the "image" label
           zTxt = "PowerBASIC rules!"                 ' it does, doesn't it..
           h    = ofnMakeFont("Times New Roman", 24, %FW_BOLD, 1, 0, 0) 'create a fancy font
           IF h THEN h = SelectObject(wParam, h)      ' and select it into device context

           GetClientRect lParam, rc                   ' get label's client size
           w = CreateSolidBrush(RGB(0, 128, 128))     ' create a brush for the background (RGB = color)
           FillRect wParam, rc, w                     ' paint the background
           IF w THEN DeleteObject w                   ' delete the brush when done
           SetBkMode wParam, %TRANSPARENT             ' set transparent text background

           OffsetRect rc, 1, 1                        ' move draw-rectangle 1 step down/right
           SetTextColor wParam, %BLACK                ' draw black text (shadow)
           DrawText wParam, zTxt, LEN(zTxt), rc, %DT_CENTER OR %DT_SINGLELINE OR %DT_VCENTER

           OffsetRect rc, -2, -2                      ' move draw-rectangle 2 steps up/left
           SetTextColor wParam, RGB(255, 255, 0)      ' draw yellow text
           DrawText wParam, zTxt, LEN(zTxt), rc, %DT_CENTER OR %DT_SINGLELINE OR %DT_VCENTER

           IF h THEN DeleteObject SelectObject(wParam, h) ' delete the fancy font object
           FUNCTION = GetStockObject(%NULL_BRUSH)     ' and make sure nothing is wiped out by Windows engine
        END IF

     CASE %WM_DESTROY
        IF otd.hFont THEN DeleteObject otd.hFont  ' clean up font handle on exit
        IF otd.oldofnComDlgProc THEN                  ' and un-subclass comdlg
           SetWindowLong GetParent(hWnd), %GWL_WNDPROC, otd.oldofnComDlgProc
        END IF

     CASE %WM_PAINT
        InvalidateRect otd.hImgLabel, BYVAL %NULL, 0 : UpdateWindow otd.hImgLabel

     CASE %WM_SIZE 'a bit tricky, this one - comdlg returns strange coordinates, so must look at controls..
        GetWindowRect otd.hTab(1), rc  ' tab dialog left
        MapWindowPoints 0, hWnd, rc, 2
        lPage = rc.nLeft

        GetWindowRect GetDlgItem(GetParent(hWnd), %stc1), rc 'a label at toolbar pos = top
        MapWindowPoints 0, hWnd, rc, 2

        GetWindowRect GetDlgItem(hWnd, %stc32), rc2      ' get %stc32's coordinates on screen = bottom
        MapWindowPoints 0, hWnd, rc2, 2                  ' map result to dialog
        rc.nBottom = ABS(rc2.nTop)                       ' bottom

        GetWindowRect GetParent(hWnd), rc2              ' get comdlg coordinates on screen
        MapWindowPoints 0, hWnd, rc2, 2                 ' map result to dialog
        rc.nLeft  = ABS(rc2.nLeft)                      ' comdlg returns some negative values..
        rc.nRight = ABS(rc2.nRight)                     ' left/right

        SendMessage otd.hTabMain, %TCM_GETITEMRECT, 0, VARPTR(rc2) 'get coordinates of a tab "button"
        rc2.nBottom = rc2.nBottom - rc2.nTop            'calculate tab "button" height

        h = rc.nTop - 22 - rc2.nBottom                  ' calculate image label height
        w = rc.nRight - 2 * rc.nLeft - 10               ' calculate image label width
        SetWindowPos otd.hImgLabel, 0, 0, 0, w, h, _    ' adjust IMAGE LABEL's size
                     %SWP_NOMOVE OR %SWP_NOZORDER
        UpdateWindow otd.hImgLabel                      ' and update for quicker redraw

        h = (rc.nBottom - rc.nTop) + rc2.nBottom + 12   ' calculate tab control's height
        SetWindowPos otd.hTabMain, 0, _                 ' adjust TAB CONTROL's pos and size
                     6, rc.nTop - 10 - rc2.nBottom,  w, h, %SWP_NOZORDER
        UpdateWindow otd.hTabMain

        h = h - rc2.nBottom - 10                        ' deduct tab header's height
        w = w - lPage                                   ' calculate tab-page dialog's width
        SetWindowPos otd.hTab(1), 0, 0, 0, w, h, _      ' adjust TAB-PAGE DIALOG's size
                     %SWP_NOMOVE OR %SWP_NOZORDER
        UpdateWindow otd.hTab(1)

        GetWindowRect otd.hListView, rc2                ' get ListView coordinates on screen
        MapWindowPoints 0, GetParent(hWnd), rc2, 2      ' map result to comdlg
        w = rc.nRight - 2 * rc2.nLeft - 2               ' calculate listview's width

        GetWindowRect GetDlgItem(otd.hTab(0), %IDOK), rc ' get Ok button's pos
        MapWindowPoints 0, otd.hTab(1), rc, 2            ' map result to dialog

        GetWindowRect otd.hListView, rc2                ' get ListView coordinates on screen
        MapWindowPoints 0, otd.hTab(1), rc2, 2          ' map result to parent dialog
        h = rc.nTop - rc2.nTop - 8                      ' calculate listview's height
        SetWindowPos otd.hListView, 0, 0, 0, w, h, _    ' adjust LISTVIEW's size
                     %SWP_NOMOVE OR %SWP_NOZORDER

        GetClientRect otd.hListView, rc                 ' get listview's client area
        ListView_SetColumnWidth otd.hListView, 0, rc.nRight * 0.3 're-adjust columns
        ListView_SetColumnWidth otd.hListView, 1, rc.nRight * 0.7
        UpdateWindow otd.hListView

 
     CASE %WM_NOTIFY                              ' a notification message is coming in
        pNhmdr = lParam                           ' lParam points to an NMHDR structure with info                     
        SELECT CASE @pNhmdr.idFrom                ' idFrom tells us what control it's from
           CASE %IDC_TAB                          ' if from tab control
              lPage = TabCtrl_GetCurSel(@pNhmdr.hwndFrom) ' see what page is calling

              SELECT CASE @pNhmdr.Code   ' see what notification is sent
                 CASE %TCN_SELCHANGING   ' lPage is old page num
                    IF lPage = 0 THEN    ' if page 1, hide comdlg controls
                       REDIM ofnCtrl(0)  ' we'll store needed handles until it's show-time again
                       EnumChildWindows GetParent(hWnd), CODEPTR(ofnEnumChildsProc), 0
                       FOR ii = 0 TO UBOUND(ofnCtrl)
                          IF ofnCtrl(ii) AND isWindowVisible(ofnCtrl(ii)) THEN
                             ShowWindow ofnCtrl(ii), %SW_HIDE
                          END IF
                       NEXT
                    ELSE
                       DIALOG SHOW STATE otd.hTab(lPage), %SW_HIDE
                    END IF

                 CASE %TCN_SELCHANGE     ' lPage is new page num
                    IF lPage = 0 THEN    ' if page 1, show comdlg controls
                       FOR ii = 0 TO UBOUND(ofnCtrl) ' use previously store handles
                          IF ofnCtrl(ii) THEN ShowWindow ofnCtrl(ii), %SW_SHOW
                       NEXT

                    ELSE
                       DIALOG SHOW STATE otd.hTab(lPage), %SW_SHOW

                    END IF
                    InvalidateRect otd.hTab(lPage), BYVAL %NULL, 0 ' refresh page after change
                    UpdateWindow otd.hTab(lPage)
                    otd.TabPg = lPage
              END SELECT
              EXIT FUNCTION

           CASE ELSE 'else it most likely comes from comdlg - see what notification is sent
              SELECT CASE @pNhmdr.Code
                 CASE %CDN_INITDONE 'is sent right after WM_INITDIALOG ..
                    'center dialog above/in parent
                    ofnCenterChild GetParent(GetParent(hWnd)), GetParent(hWnd)

                    'increase height of dropped list in file type combo (nicer, if many file types..)
                    hftCombo = GetDlgItem(GetParent(hWnd), %cmb1)     'handle, Filetype combo
                    IF hftCombo THEN                                  'if we get handle
                       GetClientRect hftCombo, rc                     'get combo's width and set new height
                       SetWindowPos hftCombo, 0, 0, 0, rc.nRight, 200, %SWP_NOMOVE OR %SWP_NOZORDER
                    END IF

                 CASE %CDN_HELP 'here we can call WinHelp API in response to help button..
                    IF TabCtrl_GetCurSel(otd.hTabMain) = 0 THEN
                       MSGBOX "Here you can show help for OpenFile dialog."
                    ELSE
                       MSGBOX "Here you can show help for Recent files."
                    END IF

              END SELECT

        END SELECT

  END SELECT
END FUNCTION

'
' Get string from ini file
'
FUNCTION ofnIniGetString(BYVAL sSection AS STRING, BYVAL sKey AS STRING, _
                         BYVAL sDefault AS STRING) AS STRING

  LOCAL RetVal AS LONG, zSection AS ASCIIZ * 80, zKey AS ASCIIZ * 80, _
        zDefault AS ASCIIZ * %MAX_PATH, zResult AS ASCIIZ * %MAX_PATH

  zSection = sSection
  zKey     = sKey
  zDefault = sDefault
  RetVal   = GetPrivateProfileString(zSection, zKey, zDefault, _
                                     zResult, SIZEOF(zResult), otd.zIni)

  IF RetVal THEN FUNCTION = TRIM$(LEFT$(zResult, RetVal))
END FUNCTION

'
' Set string to ini file
'
FUNCTION ofnIniSetString(BYVAL sSection AS STRING, BYVAL sKey AS STRING, _
                         BYVAL sStr AS STRING) AS LONG

  LOCAL zSection AS ASCIIZ * 80, zKey AS ASCIIZ * 80, zStr AS ASCIIZ * %MAX_PATH

  zSection = sSection
  zKey     = sKey
  zStr     = sStr

  FUNCTION = WritePrivateProfileString(zSection, zKey, zStr, otd.zIni)
END FUNCTION

'
' Get strings from ini file to ListView
'
SUB ofnIniToListView
  LOCAL c AS LONG, d AS LONG, hImageList AS DWORD
  LOCAL lpshfi AS SHFILEINFO, tLVI AS LV_ITEM
  LOCAL zTxt AS ASCIIZ * %MAX_PATH, zPart AS ASCIIZ * %MAX_PATH

  d = 1
  DO
     zTxt = ofnIniGetString("Recent", "File" & FORMAT$(d), "") ' read from ini file
     IF LEN(zTxt) AND ofnFileExist(zTxt) THEN
        tLVI.stateMask = %LVIS_FOCUSED
        tLVI.pszText   = VARPTR(zPart)
        tLVI.iItem     = c

        tLVI.iSubItem  = 0
        tLVI.lParam    = tLVI.iItem
        hImageList     = SHGetFileInfo(zTxt, 0, lpshfi, LEN(lpshfi), _  ' get associated icom
                         %SHGFI_ICON OR %SHGFI_SMALLICON OR %SHGFI_SYSICONINDEX)
        tLVI.iImage    = lpshfi.iIcon        ' associated icon in first column
        zPart          = ofnGetFileNamePart(zTxt) ' file name in first column 
        tLVI.mask      = %LVIF_TEXT OR %LVIF_PARAM OR %LVIF_STATE OR %LVIF_IMAGE
        ListView_InsertItem otd.hListView, tLVI

        IF c = 1 AND hImageList THEN ' we use system's imagelist, so only need to set it once
           SendMessage otd.hListView, %LVM_SETIMAGELIST, %LVSIL_SMALL, hImageList
        END IF

        tLVI.iSubItem  = 1
        zPart          = ofnGetFilePathPart(zTxt) ' file path in second column
        tLVI.mask      = %LVIF_TEXT          ' and no icon here
        ListView_SetItem otd.hListView, tLVI
        INCR c

     ELSEIF LEN(zTxt) = 0 THEN
        EXIT DO
     END IF
     INCR d
  LOOP

END SUB

'
' Some differences in old/new Win, so need to know if we're in Win2K, or later
'
FUNCTION ofnIsWin2000orLater AS LONG
  LOCAL os AS OSVERSIONINFO
  os.dwOSVersionInfoSize = SIZEOF(os)
  IF GetVersionEx(os) THEN
     FUNCTION = ((os.dwPlatformId = %VER_PLATFORM_WIN32_NT) AND _
                (os.dwMajorVersion > 4)) '5 or higher..
  END IF
END FUNCTION

'
' Create a desirable font and return its handle. Original code by Dave Navarro
' Here enhanced with proper enumeration of character set via EnumCharSet
'
FUNCTION ofnMakeFont(BYVAL sFont AS STRING, BYVAL PointSize AS LONG, _
                    BYVAL fBold AS LONG, BYVAL fItalic AS LONG, _
                    BYVAL fUnderline AS LONG, BYVAL StrikeThru AS LONG) AS LONG

  LOCAL hDC AS LONG, CharSet AS LONG, CyPixels AS LONG

  hDC = GetDC(%HWND_DESKTOP)
    CyPixels  = GetDeviceCaps(hDC, %LOGPIXELSY)
    EnumFontFamilies hDC, BYVAL STRPTR(sFont), _
                          CODEPTR(ofnMakeFontEnumCharSet), BYVAL VARPTR(CharSet)
  ReleaseDC %HWND_DESKTOP, hDC
  PointSize = 0 - (PointSize * CyPixels) \ 72

  FUNCTION = CreateFont(PointSize, 0, _  'height, width(default=0)
             0, 0, _                     'escapement(angle), orientation
             fBold, _                    'weight (%FW_DONTCARE = 0, %FW_NORMAL = 400, %FW_BOLD = 700)
             fItalic, _                  'Italic
             fUnderline, _               'Underline
             StrikeThru, _               'StrikeThru
             CharSet, %OUT_TT_PRECIS, _
             %CLIP_DEFAULT_PRECIS, %DEFAULT_QUALITY, _
             %FF_DONTCARE , BYCOPY sFont)

END FUNCTION

'
' Get type of character set - ansi, symbol.. a must for some fonts.
'
FUNCTION ofnMakeFontEnumCharSet (elf AS ENUMLOGFONT, ntm AS NEWTEXTMETRIC, _
                                 BYVAL FontType AS LONG, CharSet AS LONG) AS LONG
  CharSet = elf.elfLogFont.lfCharSet
END FUNCTION

'
' Customized GetOpenFilename procedure
' slightly different from the one PB provides us with - I have made
' sure it can handle both single sel and multiple sel.
'
FUNCTION ofnOpenFileTemplate (BYVAL hWnd AS DWORD, _           ' parent window
                              BYVAL Caption AS STRING, _       ' caption
                              Filespec AS STRING, _            ' file name
                              BYVAL InitialDir AS STRING, _    ' start directory
                              BYVAL Filter AS STRING, _        ' file type filters
                              BYVAL DefExtension AS STRING, _  ' default file type
                              Flags AS DWORD) AS LONG          ' flags

  LOCAL ofnResult     AS LONG
  LOCAL sFile         AS STRING
  LOCAL zFileTitle    AS ASCIIZ * %MAX_PATH
  LOCAL sFilter       AS STRING
  LOCAL zInitialDir   AS ASCIIZ * %MAX_PATH
  LOCAL zTitle        AS ASCIIZ * %MAX_PATH
  LOCAL zDefExt       AS ASCIIZ * %MAX_PATH
  LOCAL zOfnDlgName   AS ASCIIZ * 80
  LOCAL ofn           AS OPENFILENAME5 'use new Win2K structure!

  REPLACE "|" WITH CHR$(0) IN Filter

  IF LEN(InitialDir) = 0 THEN InitialDir = CURDIR$

  sFilter     = Filter       + CHR$(0)
  zInitialDir = InitialDir   + CHR$(0)
  zDefExt     = DefExtension + CHR$(0)
  zTitle      = Caption      + CHR$(0)
  sFile       = STRING$(16000, 0)  '<- make room for multiple files
  MID$(sFile,1) = Filespec   + CHR$(0)

  IF ofnIsWin2000orLater THEN
     ofn.lStructSize       = LEN(OPENFILENAME5) ' if Win2K or later, use full size of new structure
  ELSE
     ofn.lStructSize       = LEN(OPENFILENAME)   ' else set size to old, smaller one (76 bytes)
  END IF

  ofn.hWndOwner         = hWnd
  ofn.hInstance         = GetModuleHandle("")
  ofn.lpstrFilter       = stRPTR(sFilter)
  ofn.lpstrFile         = STRPTR(sFile)
  ofn.nMaxFile          = LEN(sFile)
  ofn.lpstrFileTitle    = VARPTR(zFileTitle)
  ofn.nMaxFileTitle     = SIZEOF(zFileTitle)
  ofn.lpstrInitialDir   = VARPTR(zInitialDir)
  ofn.Flags             = Flags
  ofn.lpfnHook          = CODEPTR(OfnHookProc)
  ofn.lpstrDefExt       = VARPTR(zDefExt)
  ofn.nFilterIndex      = 1  ' nFilterIndex decides File type combo's ListItem
  zOfnDlgName           = "OFNTEMPLATE" 
  ofn.lpTemplateName    = VARPTR(zOfnDlgName)
  IF LEN(zTitle) THEN ofn.lpstrTitle = VARPTR(zTitle)

  ' store flags in global UDT so we have access to it from tab page 2
  otd.ofnFlags           = Flags

  ofnResult = GetOpenFilename(BYVAL VARPTR(ofn)) '<- this is where OpenFile dialog is called upon!

  IF otd.TabPg THEN               ' if user pressed ok on tab-page 2,
     ofnResult = otd.TabPg        ' we must assign own return-result
     Flags     = otd.ofnFlags
     IF ofnResult > 0 AND LEN(gOfnFiles) THEN
        Filespec = gOfnFiles
     ELSE
        Filespec = ""
     END IF

  ELSE                            'else tab page 1, ordinary OpenFile response..
     Flags = ofn.Flags
     IF ofnResult > 0 THEN
        Filespec = RTRIM$(sFile, CHR$(0))
     ELSE
        Filespec = ""
     END IF
  END IF

  gOfnFiles = "" ' it's global, so make sure it's empty, now that we don't need it anymore..
  FUNCTION  = ofnResult

END FUNCTION

'
' Tab-page 2 procedure - Recent files tab
'
CALLBACK FUNCTION ofnTabPage2Proc() AS LONG
  LOCAL i AS LONG, pNmlv AS NM_LISTVIEW PTR
  LOCAL zfName AS ASCIIZ * %MAX_PATH, zPath AS ASCIIZ * %MAX_PATH

  SELECT CASE CBMSG
     CASE %WM_NOTIFY                                ' a notification message is coming in
        pNmlv = CBLPARAM                            ' lParam points to an NM_LISTVIEW structure
        SELECT CASE @pNmlv.hdr.Code
           CASE %NM_DBLCLK                          ' If double-click in ListView,
              IF @pNmlv.iItem > -1 THEN             ' only one can be selected, so grab name and path
                 gOfnFiles = ""
                 ListView_GetItemText otd.hListView, @pNmlv.iItem, 0, zfName, SIZEOF(zfName)
                 ListView_GetItemText otd.hListView, @pNmlv.iItem, 1, zPath, SIZEOF(zPath)
                 gOfnFiles = zPath + zfName

                 IF ofnFileExist (gOfnFiles) THEN                  ' make sure it exists
                    i = GetDlgItem(otd.hTab(0), %chx1)             ' look for read-only checkbox
                    IF i AND IsWindowEnabled(i) THEN               ' if it's there and enabled
                       IF SendMessage(i, %BM_GETCHECK, 0, 0) THEN  ' see if it's checked
                          otd.ofnFlags = otd.ofnFlags OR %OFN_READONLY ' set read-only flag for return
                       END IF
                    END IF

                    EndDialog otd.hTab(0), 0         ' and close dialog
                 END IF
              END IF
        END SELECT

   END SELECT

END FUNCTION

'
' Update the list of recently opened files
' If no ini file exists and path is valid, WritePrivateProfileString makes sure
' is is created. If you prefer to use the registry instead, you can still use
' this routine, but change ini calls to registry equivalents. I personally hate
' that fragile and extremely bloated registry..
'
SUB ofnWriteRecentFiles (BYVAL OpenFName AS STRING)
  LOCAL Ac AS LONG, tmpStr AS STRING
  DIM IniName() AS STRING

  Ac = 1
  DO  ' first see what already is stored
     tmpStr = ofnIniGetString("Recent", "File" & FORMAT$(Ac), "")
     IF LEN(tmpStr) THEN
        REDIM PRESERVE IniName(1 : Ac + 1) ' use a 1-based array
        IniName(Ac) = tmpStr               ' and store results
     ELSE
        EXIT DO
     END IF
     INCR Ac
  LOOP

  IF Ac > 1 THEN 'if we have an array, see if file already is there
     Ac = 0
     ARRAY SCAN IniName(), COLLATE UCASE, = OpenFName, TO Ac
     IF Ac THEN ARRAY DELETE IniName(Ac) ' if there, delete existing string
     ARRAY INSERT IniName(), OpenFName   ' and insert current one at top of array

  ELSE 'else dim array and add current file to it
     REDIM PRESERVE IniName(1 : 1)
     IniName(1) = OpenFName
  END IF

  FOR Ac = 1 TO MIN&(100, UBOUND(IniName)) ' allow how many Recent? 100 is plenty, but you decide..
     IF LEN(IniName(Ac)) THEN              ' but remember there's a 32 KB/section limit in Win95/98..
        ofnIniSetString "Recent", "File" & FORMAT$(Ac), IniName(Ac)
     ELSE
        EXIT FOR
     END IF
  NEXT Ac

END SUB

